import traitlets as tl

from aiidalab_qe_base.mixins import HasBlockers
from aiidalab_qe_base.panels.panel import PanelModel


class SettingsModel(PanelModel, HasBlockers):
    """Base model for settings models."""

    title = "Settings"
    identifier = "settings"

    include = tl.Bool(False)
    loaded_from_process = tl.Bool(False)

    def update(self):
        """Updates the model."""
        pass

    def get_model_state(self) -> dict:
        """Retrieves the current state of the model as a dictionary."""
        raise NotImplementedError()

    def set_model_state(self, parameters: dict):
        """Distributes the parameters of a loaded calculation to the model."""
        raise NotImplementedError()

    def reset(self):
        """Resets the model to present defaults."""
        pass



from __future__ import annotations

from threading import Event, Thread
from time import time

import ipywidgets as ipw
import traitlets as tl
from aiida import orm
from aiida.common.exceptions import NotExistent
from aiidalab_widgets_base import ComputationalResourcesWidget, LoadingWidget
from IPython.display import clear_output, display


class InfoBox(ipw.VBox):
    """The `InfoBox` component is used to provide additional info regarding a widget or an app."""

    def __init__(self, classes: list[str] | None = None, **kwargs):
        """`InfoBox` constructor.

        Parameters
        ----------
        `classes` : `list[str]`, optional
            One or more CSS classes.
        """
        super().__init__(**kwargs)
        self.add_class("info-box")
        for custom_classes in classes or []:
            for custom_class in custom_classes.split(" "):
                if custom_class:
                    self.add_class(custom_class)


class InAppGuide(InfoBox):
    """The `InAppGuide` is used to set up toggleable in-app guides.

    Attributes
    ----------
    `manager` : `GuideManager`
        A local reference to the global guide manager.
    `identifier` : `str`, optional
        If content `children` are not provided directly, the `identifier`
        is used to fetch the corresponding guide section from the guide
        currently loaded by the guide manager.

    Raises
    ------
    `ValueError`
        If neither content `children` or a guide section `identifier` are provided.
    """

    def __init__(
        self,
        children: list | None = None,
        guide_id: str = "",
        identifier: str = "",
        classes: list[str] | None = None,
        **kwargs,
    ):
        """`InAppGuide` constructor.

        Parameters
        ----------
        `children` : `list`, optional
            The content children of this guide section.
        `guide_id` : `str`, optional
            The associated guide id to be used in conjunction with content children.
            If none provided, the widget-based guide section will be shown for all
            guides.
        `identifier` : `str`, optional
            If content `children` are not provided directly, the `identifier`
            is used to fetch the corresponding guide section from the guide
            currently loaded by the guide manager.
        `classes` : `list[str]`, optional
            One or more CSS classes.
        """
        from aiidalab_qe.common.guide_manager import guide_manager

        self.manager = guide_manager

        super().__init__(
            classes=[
                "in-app-guide",
                *(classes or []),
            ],
            **kwargs,
        )

        if children:
            self.guide_id = guide_id
            self.children = children
            self.identifier = None
        elif identifier:
            self.guide_id = None
            self.children = []
            self.identifier = identifier
        else:
            raise ValueError("No widgets or path identifier provided")

        self.manager.observe(
            self._on_active_guide_change,
            "active_guide",
        )

        # This manual toggle call is necessary because the guide
        # may be contained in a component that was not yet rendered
        # when a guide was selected.
        self._on_active_guide_change(None)

    def _on_active_guide_change(self, _):
        self._update_contents()
        self._toggle_guide()

    def _update_contents(self):
        """Update the contents of the guide section."""
        if not self.identifier:
            return
        html = self.manager.get_guide_section_by_id(self.identifier)
        self.children = [ipw.HTML(str(html))] if html else []

    def _toggle_guide(self):
        """Toggle the visibility of the guide section."""
        self.layout.display = (
            "flex"
            if self.children
            and (
                # file-based guide section
                (self.identifier and self.manager.has_guide)
                # widget-based guide section shown for all guides
                or (not self.guide_id and self.manager.has_guide)
                # widget-based guide section shown for a specific guide
                or self.guide_id == self.manager.active_guide
            )
            else "none"
        )


class LinkButton(ipw.HTML):
    disabled = tl.Bool(False)

    def __init__(
        self,
        description=None,
        link="",
        in_place=False,
        class_="",
        style_="",
        icon="",
        tooltip="",
        disabled=False,
        **kwargs,
    ):
        html = f"""
            <a
                role="button"
                href="{link}"
                title="{tooltip or description}"
                target="{"_self" if in_place else "_blank"}"
                style="cursor: default; {style_}"
            >
        """
        if icon:
            html += f"<i class='fa fa-{icon}'></i>"

        html += f"{description}</a>"

        super().__init__(value=html, **kwargs)

        self.add_class("jupyter-button")
        self.add_class("widget-button")
        self.add_class("link-button")
        if class_:
            self.add_class(class_)

        self.disabled = disabled

    @tl.observe("disabled")
    def _on_disabled(self, change):
        if change["new"]:
            self.add_class("disabled")
        else:
            self.remove_class("disabled")


class HBoxWithUnits(ipw.HBox):
    def __init__(self, widget: ipw.ValueWidget, units: str, **kwargs):
        super().__init__(
            children=[
                widget,
                ipw.HTML(units),
            ],
            layout={
                "align_items": "center",
                "grid_gap": "2px",
            }
            | kwargs.pop("layout", {}),
            **kwargs,
        )
        self.add_class("hbox-with-units")


class LazyLoader(ipw.VBox):
    identifier = "widget"

    def __init__(self, widget_class, widget_kwargs=None, **kwargs):
        super().__init__(
            children=[LoadingWidget(f"Loading {self.identifier}")],
            **kwargs,
        )

        self._widget_class = widget_class
        self._widget_kwargs = widget_kwargs or {}

        self.rendered = False

    def set_widget_kwargs(self, kwargs):
        self._widget_kwargs = kwargs

    def render(self):
        if self.rendered:
            return
        self.widget = self._widget_class(**self._widget_kwargs)
        self.children = [self.widget]
        self.rendered = True


class ProgressBar(ipw.HBox):
    class AnimationRate(float):
        pass

    description = tl.Unicode()
    value = tl.Union([tl.Float(), tl.Instance(AnimationRate)])
    bar_style = tl.Unicode()

    _animation_rate = tl.Float()

    def __init__(self, description_layout=None, *args, **kwargs):
        if description_layout is None:
            description_layout = ipw.Layout(width="auto", flex="2 1 auto")

        self._label = ipw.Label(layout=description_layout)
        self._progress_bar = ipw.FloatProgress(
            min=0, max=1.0, layout=ipw.Layout(width="auto", flex="1 1 auto")
        )

        tl.link((self, "description"), (self._label, "value"))
        tl.link((self, "bar_style"), (self._progress_bar, "bar_style"))

        self._animate_stop_event = Event()
        self._animate_thread = None

        super().__init__([self._label, self._progress_bar], *args, **kwargs)

    def _animate(self, refresh_rate=0.01):
        v0 = self._progress_bar.value
        t0 = time()

        while not self._animate_stop_event.wait(refresh_rate):
            self._progress_bar.value = (v0 + (time() - t0) * self._animation_rate) % 1.0

    def _start_animate(self):
        if self._animate_thread is not None:
            raise RuntimeError("Cannot start animation more than once!")

        self._animate_thread = Thread(target=self._animate)
        self._animate_thread.start()

    def _stop_animate(self):
        self._animate_stop_event.set()
        self._animate_thread.join()
        self._animate_stop_event.clear()
        self._animate_thread = None

    @tl.default("_animation_rate")
    def _default_animation_rate(self):
        return 0

    @tl.observe("_animation_rate")
    def _observe_animation_rate(self, change):
        if change["new"] and not change["old"]:
            self._start_animate()
        elif not change["new"] and change["old"]:
            self._stop_animate()

    @tl.validate("value")
    def _validate_value(self, proposal):
        if isinstance(proposal["value"], self.AnimationRate):
            if proposal["value"] < 0:
                raise tl.TraitError("The animation rate must be non-negative.")

        elif not 0 <= proposal["value"] <= 1.0:
            raise tl.TraitError("The value must be between 0 and 1.0.")

        return proposal["value"]

    @tl.observe("value")
    def _observe_value(self, change):
        if isinstance(change["new"], self.AnimationRate):
            self._animation_rate = change["new"]
        else:
            self._animation_rate = 0
            self._progress_bar.value = change["new"]


class QEAppComputationalResourcesWidget(ipw.VBox):
    value = tl.Unicode(allow_none=True)
    nodes = tl.Int(default_value=1)
    cpus = tl.Int(default_value=1)

    def __init__(self, **kwargs):
        """Widget to setup the compute resources, which include the code,
        the number of nodes and the number of cpus.
        """
        self.code_selection = ComputationalResourcesWidget(
            description=kwargs.pop("description", None),
            default_calc_job_plugin=kwargs.pop("default_calc_job_plugin", None),
            include_setup_widget=False,
            fetch_codes=False,
            **kwargs,
        )
        self.code_selection.layout.width = "80%"

        self.num_nodes = ipw.BoundedIntText(
            value=1,
            step=1,
            min=1,
            max=1000,
            description="Nodes",
        )

        self.num_cpus = ipw.BoundedIntText(
            value=1,
            step=1,
            min=1,
            description="CPUs",
        )

        self.btn_setup_resource_detail = ipw.ToggleButton(description="More")
        self.btn_setup_resource_detail.observe(
            self._setup_resource_detail,
            "value",
        )

        self._setup_resource_detail_output = ipw.Output(layout={"width": "500px"})

        # combine code, nodes and cpus
        children = [
            ipw.HBox(
                children=[
                    self.code_selection,
                    self.num_nodes,
                    self.num_cpus,
                    self.btn_setup_resource_detail,
                ]
            ),
            self._setup_resource_detail_output,
        ]
        super().__init__(children=children, **kwargs)

        self.resource_detail = ResourceDetailSettings()

        tl.dlink(
            (self.num_cpus, "value"),
            (self.resource_detail.ntasks_per_node, "value"),
        )

        tl.link((self.code_selection, "value"), (self, "value"))

    def update_resources(self, change):
        if change["new"]:
            try:
                computer = orm.load_code(change["new"]).computer
            except NotExistent:
                computer = None
            self.set_resource_defaults(computer)

    def set_resource_defaults(self, computer=None):
        if computer is None:
            self.num_nodes.disabled = True
            self.num_nodes.value = 1
            self.num_cpus.max = 1
            self.num_cpus.value = 1
            self.num_cpus.description = "CPUs"
        else:
            default_mpiprocs = computer.get_default_mpiprocs_per_machine()
            self.num_nodes.disabled = (
                True if computer.hostname == "localhost" else False
            )
            self.num_cpus.max = default_mpiprocs
            self.num_cpus.value = (
                1 if computer.hostname == "localhost" else default_mpiprocs
            )
            self.num_cpus.description = "CPUs"

    @property
    def parameters(self):
        return self.get_parameters()

    def get_parameters(self):
        """Return the parameters."""
        parameters = {
            "code": self.code_selection.value,
            "nodes": self.num_nodes.value,
            "cpus": self.num_cpus.value,
        }
        parameters.update(self.resource_detail.parameters)
        return parameters

    @parameters.setter
    def parameters(self, parameters):
        self.set_parameters(parameters)

    def set_parameters(self, parameters):
        """Set the parameters."""
        self.code_selection.value = parameters["code"]
        if "nodes" in parameters:
            self.num_nodes.value = parameters["nodes"]
        if "cpus" in parameters:
            self.num_cpus.value = parameters["cpus"]
        if "ntasks_per_node" in parameters:
            self.resource_detail.ntasks_per_node.value = parameters["ntasks_per_node"]
        if "cpus_per_task" in parameters:
            self.resource_detail.cpus_per_task.value = parameters["cpus_per_task"]
        if "max_wallclock_seconds" in parameters:
            self.resource_detail.max_wallclock_seconds.value = parameters[
                "max_wallclock_seconds"
            ]

    def _setup_resource_detail(self, _=None):
        with self._setup_resource_detail_output:
            clear_output()
            if self.btn_setup_resource_detail.value:
                self._setup_resource_detail_output.layout = {
                    "width": "500px",
                    "border": "1px solid gray",
                }

                children = [
                    self.resource_detail,
                ]
                display(*children)
            else:
                self._setup_resource_detail_output.layout = {
                    "width": "500px",
                    "border": "none",
                }


class ResourceDetailSettings(ipw.VBox):
    """Widget for setting the Resource detail."""

    prompt = ipw.HTML(
        """<div style="line-height:120%; padding-top:0px">
        <p style="padding-bottom:10px">
        Specify the parameters for the scheduler (only for advanced user). <br>
        These should be specified accordingly to the computer where the code will run.
        </p></div>"""
    )

    def __init__(self, **kwargs):
        self.ntasks_per_node = ipw.BoundedIntText(
            value=1,
            step=1,
            min=1,
            max=1000,
            description="ntasks-per-node",
            style={"description_width": "100px"},
        )
        self.cpus_per_task = ipw.BoundedIntText(
            value=1,
            step=1,
            min=1,
            description="cpus-per-task",
            style={"description_width": "100px"},
        )
        self.max_wallclock_seconds = ipw.BoundedIntText(
            value=3600 * 12,
            step=3600,
            min=60 * 10,
            max=3600 * 24,
            description="max seconds",
            style={"description_width": "100px"},
        )
        super().__init__(
            children=[
                self.prompt,
                self.ntasks_per_node,
                self.cpus_per_task,
                self.max_wallclock_seconds,
            ],
            **kwargs,
        )

    @property
    def parameters(self):
        return self.get_parameters()

    def get_parameters(self):
        """Return the parameters."""
        return {
            "ntasks_per_node": self.ntasks_per_node.value,
            "cpus_per_task": self.cpus_per_task.value,
            "max_wallclock_seconds": self.max_wallclock_seconds.value,
        }

    @parameters.setter
    def parameters(self, parameters):
        self.ntasks_per_node.value = parameters.get("ntasks_per_node", 1)
        self.cpus_per_task.value = parameters.get("cpus_per_task", 1)
        self.max_wallclock_seconds.value = parameters.get(
            "max_wallclock_seconds", 3600 * 12
        )

    def reset(self):
        """Reset the settings."""
        self.ntasks_per_node.value = 1
        self.cpus_per_task.value = 1
        self.max_wallclock_seconds.value = 3600 * 12


class ParallelizationSettings(ipw.VBox):
    """Widget for setting the parallelization settings."""

    prompt = ipw.HTML(
        """<div style="line-height:120%; padding-top:0px">
        <p style="padding-bottom:10px">
        Specify the number of k-points pools for the pw.x calculations (only for advanced user).
        </p></div>"""
    )

    def __init__(self, **kwargs):
        extra = {
            "style": {"description_width": "150px"},
            "layout": {"min_width": "180px"},
        }
        self.npool = ipw.BoundedIntText(
            value=1,
            step=1,
            min=1,
            max=128,
            description="Number of k-pools",
            **extra,
        )
        self.override = ipw.Checkbox(
            description="",
            indent=False,
            value=False,
            layout=ipw.Layout(max_width="20px"),
        )
        ipw.dlink(
            (self.override, "value"),
            (self.npool.layout, "display"),
            lambda override: "block" if override else "none",
        )

        super().__init__(
            children=[
                ipw.HBox(
                    children=[
                        self.override,
                        self.prompt,
                        self.npool,
                    ],
                    layout=ipw.Layout(justify_content="flex-start"),
                ),
            ],
            **kwargs,
        )

        # set the default visibility of the widget
        self.npool.layout.display = "none"

    def reset(self):
        self.npool.value = 1


class PwCodeResourceSetupWidget(QEAppComputationalResourcesWidget):
    """ComputationalResources Widget for the pw.x calculation."""

    nodes = tl.Int(default_value=1)

    def __init__(self, **kwargs):
        # By definition, npool must be a divisor of the total number of k-points
        # thus we can not set a default value here, or from the computer.
        self.parallelization = ParallelizationSettings()
        super().__init__(**kwargs)
        # add nodes and cpus into the children of the widget
        self.children += (self.parallelization,)

    def get_parallelization(self):
        """Return the parallelization settings."""
        parallelization = (
            {"npool": self.parallelization.npool.value}
            if self.parallelization.override.value
            else {}
        )
        return parallelization

    def set_parallelization(self, parallelization):
        """Set the parallelization settings."""
        if "npool" in parallelization:
            self.parallelization.override.value = True
            self.parallelization.npool.value = parallelization["npool"]

    def get_parameters(self):
        """Return the parameters."""
        parameters = super().get_parameters()
        parameters.update({"parallelization": self.get_parallelization()})
        return parameters

    def set_parameters(self, parameters):
        """Set the parameters."""
        super().set_parameters(parameters)
        if "parallelization" in parameters:
            self.set_parallelization(parameters["parallelization"])



import traitlets as tl

from aiidalab_qe_base.models import CodeModel
from aiidalab_qe_base.panels.resources import ResourceSettingsModel


class PluginResourceSettingsModel(ResourceSettingsModel):
    """Base model for plugin resource setting models."""

    dependencies = [
        "global.global_codes",
    ]

    override = tl.Bool(False)

    def add_model(self, identifier: str, model: CodeModel):
        super().add_model(identifier, model)
        model.activate()

    def update(self):
        """Updates the code models from the global resources.

        Skips synchronization with global resources if the user has chosen to override
        the resources for the plugin codes.
        """
        if self.override:
            return
        for _, code_model in self.get_models():
            model_key = code_model.default_calc_job_plugin.replace(".", "__")
            if model_key in self.global_codes:
                code_resources: dict = self.global_codes[model_key]  # type: ignore
                code_model.set_model_state(code_resources)

    def get_model_state(self):
        return {
            "override": self.override,
            **super().get_model_state(),
        }

    def set_model_state(self, parameters: dict):
        self.override = parameters.get("override", False)
        super().set_model_state(parameters)

    def _link_model(self, model: CodeModel):
        tl.link(
            (self, "override"),
            (model, "override"),
        )



import sys
import typing as t
from datetime import datetime

import traitlets as tl
from aiida import orm
from dateutil.relativedelta import relativedelta

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class HasTraits(tl.HasTraits):
    """Wrapper on `tl.HasTraits` for improved static type checking."""

    # For IDE type checking
    # Type checking struggles with `traitlets.HasTraits`, which inherits
    # `traitlets.HasDescriptors`, which in turn defines `__new__(...) -> t.Any`
    def __new__(cls, *args: t.Any, **kwargs: t.Any) -> Self:
        return super().__new__(cls, *args, **kwargs)


def set_component_resources(component, code_info):
    """Set the resources for a given component based on the code info."""
    # Ensure code_info is not None or empty
    # ? XXX: from jyu, need to pop a warning to plugin developer or what?
    if code_info:
        code: orm.Code = code_info["code"]
        if code.computer.scheduler_type == "hyperqueue":
            component.metadata.options.resources = {
                "num_cpus": code_info["nodes"]
                * code_info["ntasks_per_node"]
                * code_info["cpus_per_task"]
            }
        else:
            # XXX: jyu should properly deal with None type of scheduler_type which can
            # be "core.direct" (will be replaced by hyperqueue) and "core.slurm" ...
            component.metadata.options.resources = {
                "num_machines": code_info["nodes"],
                "num_mpiprocs_per_machine": code_info["ntasks_per_node"],
                "num_cores_per_mpiproc": code_info["cpus_per_task"],
            }

        max_wallclock_seconds = code_info["max_wallclock_seconds"]
        component.metadata.options["max_wallclock_seconds"] = max_wallclock_seconds

        if "parallelization" in code_info:
            component.parallelization = orm.Dict(dict=code_info["parallelization"])


def enable_pencil_decomposition(component):
    """Enable the pencil decomposition for the given component."""

    component.settings = orm.Dict({"CMDLINE": ["-pd", ".true."]})


def shallow_copy_nested_dict(d):
    """Recursively copies only the dictionary structure but keeps value references."""
    if isinstance(d, dict):
        return {key: shallow_copy_nested_dict(value) for key, value in d.items()}
    return d


def format_time(time: datetime):
    return time.strftime("%Y-%m-%d %H:%M:%S")


def relative_time(time: datetime):
    # TODO consider using humanize or arrow libraries for this
    now = datetime.now(time.tzinfo)
    delta = relativedelta(now, time)
    if delta.years > 0:
        return f"{delta.years} year{'s' if delta.years > 1 else ''} ago"
    elif delta.months > 0:
        return f"{delta.months} month{'s' if delta.months > 1 else ''} ago"
    elif delta.days > 0:
        return f"{delta.days} day{'s' if delta.days > 1 else ''} ago"
    elif delta.hours > 0:
        return f"{delta.hours} hour{'s' if delta.hours > 1 else ''} ago"
    elif delta.minutes > 0:
        return f"{delta.minutes} minute{'s' if delta.minutes > 1 else ''} ago"
    else:
        return f"{delta.seconds} second{'s' if delta.seconds > 1 else ''} ago"



import typing as t

import ipywidgets as ipw

from aiidalab_qe_base.models import CodeModel
from aiidalab_qe_base.panels.resources import ResourceSettingsPanel
from aiidalab_qe_base import widgets

from .model import PluginResourceSettingsModel

PRSM = t.TypeVar("PRSM", bound=PluginResourceSettingsModel)


class PluginResourceSettingsPanel(ResourceSettingsPanel[PRSM]):
    """Base class for plugin resource setting panels."""

    def __init__(self, model: PRSM, **kwargs):
        super().__init__(model, **kwargs)

        self._model.observe(
            self._on_global_codes_change,
            "global_codes",
        )
        self._model.observe(
            self._on_override_change,
            "override",
        )

    def render(self):
        if self.rendered:
            return

        self.override_help = ipw.HTML(
            "Click to override the resource settings for this plugin."
        )
        self.override = ipw.Checkbox(
            description="",
            indent=False,
            layout=ipw.Layout(max_width="3%"),
        )
        ipw.link(
            (self._model, "override"),
            (self.override, "value"),
        )

        self.children = [
            ipw.HBox(
                children=[
                    self.override,
                    self.override_help,
                ]
            ),
            self.code_widgets_container,
        ]

        self.rendered = True

        # Render any active codes
        for _, code_model in self._model.get_models():
            if code_model.is_active:
                self._toggle_code(code_model)

    def _on_global_codes_change(self, _):
        self._model.update()

    def _on_override_change(self, _):
        self._model.update()

    def _render_code_widget(
        self,
        code_model: CodeModel,
        code_widget: widgets.QEAppComputationalResourcesWidget,
    ):
        super()._render_code_widget(code_model, code_widget)
        self._link_override_to_widget_disable(code_model, code_widget)

    def _link_override_to_widget_disable(
        self,
        code_model: CodeModel,
        code_widget: widgets.QEAppComputationalResourcesWidget,
    ):
        """Links the override attribute of the code model to the disable attribute
        of subwidgets of the code widget."""
        ipw.dlink(
            (code_model, "override"),
            (code_widget.code_selection.code_select_dropdown, "disabled"),
            lambda override: not override,
        )
        ipw.dlink(
            (code_model, "override"),
            (code_widget.num_cpus, "disabled"),
            lambda override: not override,
        )
        ipw.dlink(
            (code_model, "override"),
            (code_widget.num_nodes, "disabled"),
            lambda override: not override,
        )
        ipw.dlink(
            (code_model, "override"),
            (code_widget.btn_setup_resource_detail, "disabled"),
            lambda override: not override,
        )
        if isinstance(code_widget, widgets.PwCodeResourceSetupWidget):
            ipw.dlink(
                (code_model, "override"),
                (code_widget.parallelization.override, "disabled"),
                lambda override: not override,
            )
            ipw.dlink(
                (code_model, "override"),
                (code_widget.parallelization.npool, "disabled"),
                lambda override: not override,
            )



from .table_widget import TableWidget
from .widgets import (
    HBoxWithUnits,
    InAppGuide,
    InfoBox,
    LazyLoader,
    LinkButton,
    ParallelizationSettings,
    ProgressBar,
    PwCodeResourceSetupWidget,
    QEAppComputationalResourcesWidget,
    ResourceDetailSettings,
)

__all__ = [
    "TableWidget",
    "HBoxWithUnits",
    "InAppGuide",
    "InfoBox",
    "LazyLoader",
    "LinkButton",
    "ParallelizationSettings",
    "ProgressBar",
    "PwCodeResourceSetupWidget",
    "QEAppComputationalResourcesWidget",
    "ResourceDetailSettings",
]



import ipywidgets as ipw


class PluginOutline(ipw.HBox):
    title = "Outline"
    description = ""

    def __init__(self, **kwargs):
        self.include = ipw.Checkbox(
            description=self.title,
            indent=False,
            style={"description_width": "initial"},
        )

        super().__init__(
            children=[
                self.include,
                ipw.HTML(f"""
                    <div style="line-height: 140%; padding-top: 0px; padding-bottom: 5px">
                        {self.description}
                    </div>
                """),
            ],
            layout=ipw.Layout(width="fit-content"),
            **kwargs,
        )



from pathlib import Path

import traitlets as tl
from anywidget import AnyWidget


class TableWidget(AnyWidget):
    _esm = Path(__file__).parent / "table_widget.js"
    _css = Path(__file__).parent / "table_widget.css"
    data = tl.List().tag(sync=True)
    selected_rows = tl.List().tag(sync=True)



import sys
import typing as t

import ipywidgets as ipw
from aiidalab_widgets_base import LoadingWidget

from .model import PanelModel

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


PM = t.TypeVar("PM", bound=PanelModel)


class Panel(ipw.VBox, t.Generic[PM]):
    """Base class for all panels."""

    rendered = False
    _loading_message = "Loading {identifier} panel"

    # For IDE type checking
    # Type checking struggles with `traitlets.HasTraits`, which inherits
    # `traitlets.HasDescriptors`, which in turn defines `__new__(...) -> t.Any`
    def __new__(cls, *args: t.Any, **kwargs: t.Any) -> Self:
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, model: PM, **kwargs):
        loading_message = self._loading_message.format(identifier=model.identifier)
        loading_message = loading_message.replace("_", " ")
        self.loading_message: LoadingWidget = LoadingWidget(loading_message)
        super().__init__(children=[self.loading_message], **kwargs)
        self._model = model

    def render(self):
        raise NotImplementedError()



import typing as t

import ipywidgets as ipw
from aiidalab_widgets_base import LoadingWidget

from aiidalab_qe_base.models import CodeModel
from aiidalab_qe_base.widgets import (
    PwCodeResourceSetupWidget,
    QEAppComputationalResourcesWidget,
)

from ..settings import SettingsPanel
from .model import ResourceSettingsModel

RSM = t.TypeVar("RSM", bound=ResourceSettingsModel)


class ResourceSettingsPanel(SettingsPanel[RSM]):
    """Base class for resource setting panels."""

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)
        self.code_widgets: dict[str, QEAppComputationalResourcesWidget] = {}
        self.code_widgets_container = ipw.VBox()

    def register_code_trait_callbacks(self, code_model: CodeModel):
        """Registers event handlers on code model traits."""
        if code_model.default_calc_job_plugin == "quantumespresso.pw":
            code_model.observe(
                self._on_code_resource_change,
                [
                    "parallelization_override",
                    "npool",
                ],
            )
        code_model.observe(
            self._on_code_resource_change,
            [
                "selected",
                "num_cpus",
                "num_nodes",
                "ntasks_per_node",
                "cpus_per_task",
                "max_wallclock_seconds",
            ],
        )

    def _on_code_resource_change(self, _):
        pass

    def _on_code_options_change(self, change: dict):
        widget: ipw.Dropdown = change["owner"]
        widget.disabled = not widget.options

    def _toggle_code(self, code_model: CodeModel):
        if not self.rendered:
            return
        if not code_model.is_rendered:
            loading_message = LoadingWidget(f"Loading {code_model.name} code")
            self.code_widgets_container.children += (loading_message,)
        if code_model.name not in self.code_widgets:
            code_widget = code_model.code_widget_class(
                description=code_model.description,
                default_calc_job_plugin=code_model.default_calc_job_plugin,
            )
            self.code_widgets[code_model.name] = code_widget
        else:
            code_widget = self.code_widgets[code_model.name]
        if not code_model.is_rendered:
            code_widget.observe(
                code_widget.update_resources,
                "value",
            )
            self._render_code_widget(code_model, code_widget)

    def _render_code_widget(
        self,
        code_model: CodeModel,
        code_widget: QEAppComputationalResourcesWidget,
    ):
        ipw.dlink(
            (code_model, "options"),
            (code_widget.code_selection.code_select_dropdown, "options"),
        )
        ipw.link(
            (code_model, "warning"),
            (code_widget.code_selection.output, "value"),
        )
        ipw.link(
            (code_model, "selected"),
            (code_widget, "value"),
        )
        ipw.link(
            (code_model, "num_cpus"),
            (code_widget.num_cpus, "value"),
        )
        ipw.link(
            (code_model, "num_nodes"),
            (code_widget.num_nodes, "value"),
        )
        ipw.link(
            (code_model, "ntasks_per_node"),
            (code_widget.resource_detail.ntasks_per_node, "value"),
        )
        ipw.link(
            (code_model, "cpus_per_task"),
            (code_widget.resource_detail.cpus_per_task, "value"),
        )
        ipw.link(
            (code_model, "max_wallclock_seconds"),
            (code_widget.resource_detail.max_wallclock_seconds, "value"),
        )
        if isinstance(code_widget, PwCodeResourceSetupWidget):
            ipw.link(
                (code_model, "parallelization_override"),
                (code_widget.parallelization.override, "value"),
            )
            ipw.link(
                (code_model, "npool"),
                (code_widget.parallelization.npool, "value"),
            )
        code_widget.code_selection.code_select_dropdown.observe(
            self._on_code_options_change,
            "options",
        )
        code_widgets = self.code_widgets_container.children[:-1]  # type: ignore
        self.code_widgets_container.children = [*code_widgets, code_widget]
        code_model.is_rendered = True



import typing as t

from aiidalab_qe_base.mixins import Confirmable

from ..settings import SettingsModel


class ConfigurationSettingsModel(SettingsModel, Confirmable):
    """Base model for configuration settings models."""

    title = "Configuration"
    identifier = "configuration"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._defaults: dict[str, t.Any] = {}

    def update(self, specific=""):
        """Updates the model.

        Parameters
        ----------
        `specific` : `str`, optional
            If provided, specifies the level of update.
        """
        pass

    def _get_default(self, trait):
        return self._defaults.get(trait, self.traits()[trait].default_value)



from __future__ import annotations

import traitlets as tl
from aiida import orm
from aiida.common.extendeddicts import AttributeDict

from aiidalab_qe_base.mixins import HasProcess

from ..settings import SettingsModel


class ResultsModel(SettingsModel, HasProcess):
    """Base model for results models."""

    title = "Results"
    identifier = "results"

    process_status_notification = tl.Unicode("")

    _this_process_label = ""
    _this_process_uuid = None

    auto_render = False
    _completed_process = False

    CSS_MAP = {
        "finished": "success",
        "failed": "danger",
        "excepted": "danger",
        "killed": "danger",
        "queued": "warning",
        "running": "info",
        "created": "info",
    }

    @property
    def include(self):
        return self.identifier in self.properties

    @property
    def has_results(self):
        node = self.fetch_child_process_node()
        return node and node.is_finished_ok

    def update(self):
        self.auto_render = self.has_results

    def update_process_status_notification(self):
        if self._completed_process:
            self.process_status_notification = ""
            return
        status = self._get_child_process_status()
        self.process_status_notification = status
        if "success" in status:
            self._completed_process = True

    def fetch_child_process_node(self, which="this") -> orm.ProcessNode | None:
        if not self.process_uuid:
            return
        which = which.lower()
        uuid = getattr(self, f"_{which}_process_uuid")
        label = getattr(self, f"_{which}_process_label")
        if not uuid:
            root = self.fetch_process_node()
            child = next((c for c in root.called if c.process_label == label), None)
            uuid = child.uuid if child else None
        return orm.load_node(uuid) if uuid else None  # type: ignore

    def save_state(self):
        """Saves the current state of the model to the AiiDA database."""
        node = self.fetch_process_node()
        results = node.base.extras.get("results", {})
        results[self.identifier] = self.get_model_state()
        node.base.extras.set("results", results)

    def load_state(self):
        """Loads the state of the model from the AiiDA database."""
        node = self.fetch_process_node()
        results = node.base.extras.get("results", {})
        if self.identifier in results:
            self.set_model_state(results[self.identifier])

    def _get_child_process_status(self, which="this"):
        state, exit_message = self._get_child_state_and_exit_message(which)
        if state == "waiting":
            state = "running"
        status = state.upper()
        if exit_message:
            status = f"{status} ({exit_message})"
        label = "Status" if which == "this" else f"{which.capitalize()} status"
        alert_class = f"alert-{self.CSS_MAP.get(state, 'info')}"
        return f"""
            <div class="alert {alert_class}" style="padding: 5px 10px;">
                <b>{label}:</b> {status}
            </div>
        """

    def _get_child_state_and_exit_message(self, which="this"):
        if not (
            (node := self.fetch_child_process_node(which))
            and hasattr(node, "process_state")
            and node.process_state
        ):
            return "queued", None
        if node.is_failed:
            return "failed", node.exit_message
        return node.process_state.value, None

    def _get_child_outputs(self, which="this"):
        if not (node := self.fetch_child_process_node(which)):
            outputs = super().outputs
            child = which if which != "this" else self.identifier
            return getattr(outputs, child) if child in outputs else AttributeDict({})
        return AttributeDict({key: getattr(node.outputs, key) for key in node.outputs})



import typing as t

import ipywidgets as ipw

from aiidalab_qe_base.widgets import InAppGuide

from ..panel import Panel
from .model import ResultsModel

RM = t.TypeVar("RM", bound=ResultsModel)


class ResultsPanel(Panel[RM]):
    """Base class for all the result panels.

    The base class has a method to load the result of the calculation.
    And a show method to display it in the panel.
    It has a update method to update the result in the panel.
    """

    _loading_message = "Loading {identifier} results"

    def __init__(self, model: RM, **kwargs):
        super().__init__(model=model, **kwargs)
        self._model.observe(
            self._on_process_change,
            "process_uuid",
        )
        self._model.observe(
            self._on_monitor_counter_change,
            "monitor_counter",
        )

    def render(self):
        if self.rendered:
            if self._model.identifier == "structure":
                self._render()
            return

        if not self._model.has_process:
            return

        self.guide = InAppGuide(
            identifier=f"{self._model.identifier}-results",
            classes=["results-panel-guide"],
        )
        self.save_state_button = ipw.Button(
            description="Save state",
            tooltip="Save the current visualization settings",
            button_style="primary",
            icon="save",
        )
        self.save_state_button.on_click(self._save_state)
        self.load_state_button = ipw.Button(
            description="Load state",
            tooltip="Load previously saved visualization settings",
            button_style="primary",
            icon="download",
        )
        self.load_state_button.on_click(self._load_state)
        self.state_buttons = ipw.HBox(
            children=[self.save_state_button, self.load_state_button],
        )

        self.results_container = ipw.VBox()

        if self._model.auto_render:
            self.children = [
                self.guide,
                self.results_container,
            ]
            self._load_results()
        else:
            children = [self.guide]
            if (
                self._model.identifier != "structure"
                or "relax" in self._model.properties
            ):
                children.append(self._get_controls_section())
            children.append(self.results_container)
            self.children = children
            if self._model.identifier == "structure":
                self._load_results()

        self.rendered = True

    def _on_process_change(self, _):
        self._model.update()

    def _on_monitor_counter_change(self, _):
        self._model.update_process_status_notification()

    def _on_load_results_click(self, _):
        self.load_controls.children = []
        self._load_results()

    def _load_results(self):
        self.results_container.children = [self.loading_message]
        self._render()
        self._post_render()

    def _get_controls_section(self) -> ipw.VBox:
        self.process_status_notification = ipw.HTML()
        ipw.dlink(
            (self._model, "process_status_notification"),
            (self.process_status_notification, "value"),
        )

        self.load_results_button = ipw.Button(
            description="Load results",
            button_style="warning",
            tooltip="Load the results",
            icon="refresh",
        )
        ipw.dlink(
            (self._model, "monitor_counter"),
            (self.load_results_button, "disabled"),
            lambda _: not self._model.has_results,
        )
        self.load_results_button.on_click(self._on_load_results_click)

        self.load_controls = ipw.HBox(
            children=[]
            if self._model.auto_render or self._model.identifier == "structure"
            else [
                self.load_results_button,
                ipw.HTML("""
                    <div style="margin-left: 10px">
                        <b>Note:</b> Load time may vary depending on the size of the
                        calculation
                    </div>
                """),
            ]
        )

        return ipw.VBox(
            children=[
                self.process_status_notification,
                self.load_controls,
            ]
        )

    def _render(self):
        raise NotImplementedError()

    def _post_render(self):
        pass

    def _save_state(self, _=None):
        """Save the current state of the results panel."""
        self._model.save_state()

    def _load_state(self, _=None):
        """Load a previously saved state of the results panel."""
        self._model.load_state()



from __future__ import annotations

import typing as t

import traitlets as tl
from aiida import orm
from aiida.common.exceptions import NotExistent
from aiida_quantumespresso.data.hubbard_structure import HubbardStructureData

from .models import Model

from .utils import HasTraits

T = t.TypeVar("T")


class HasInputStructure(HasTraits):
    input_structure = tl.Union(
        [
            tl.Instance(orm.StructureData),
            tl.Instance(HubbardStructureData),
        ],
        allow_none=True,
    )

    @property
    def has_structure(self):
        return self.input_structure is not None

    @property
    def has_pbc(self):
        return self.has_structure and any(self.input_structure.pbc)

    @property
    def has_tags(self):
        return self.has_structure and any(
            not kind_name.isalpha()
            for kind_name in self.input_structure.get_kind_names()
        )


class HasModels(t.Generic[T]):
    def __init__(self):
        self._models: dict[str, T] = {}

    def has_model(self, identifier):
        return identifier in self._models

    def add_model(self, identifier: str, model: T):
        self._models[identifier] = model
        self._link_model(model)

    def add_models(self, models: dict[str, T]):
        for identifier, model in models.items():
            self.add_model(identifier, model)

    def get_model(self, identifier: str) -> T:
        keys = identifier.split(".", 1)
        if self.has_model(keys[0]):
            if len(keys) == 1:
                return self._models[identifier]
            else:
                sub_model = self._models[keys[0]]
                if isinstance(sub_model, HasModels):
                    return sub_model.get_model(keys[1])
                raise TypeError(
                    f"Model with identifier '{identifier}' does not have sub-models."
                )
        raise KeyError(f"Model with identifier '{identifier}' not found.")

    def get_models(self) -> t.Iterable[tuple[str, T]]:
        return self._models.items()

    def _link_model(self, model: T):
        if not hasattr(model, "dependencies"):
            return
        if isinstance(model, HasBlockers):
            tl.dlink(
                (model, "blockers"),
                (self, "blockers"),
            )
        if isinstance(model, Model):
            for dependency in model.dependencies:
                dependency_parts = dependency.rsplit(".", 1)
                if len(dependency_parts) == 1:  # from parent
                    target_model = self
                    trait = dependency
                else:  # from sibling
                    sibling, trait = dependency_parts
                    target_model = self.get_model(sibling)
                tl.dlink(
                    (target_model, trait),
                    (model, trait),
                )
        else:
            raise TypeError(f"Model linking is not supported for {type(model)}.")


class HasProcess(HasTraits):
    process_uuid = tl.Unicode(None, allow_none=True)
    monitor_counter = tl.Int(0)  # used for continuous updates

    @property
    def has_process(self):
        return self.fetch_process_node() is not None

    @property
    def inputs(self):
        process_node = self.fetch_process_node()
        return process_node.inputs if process_node else []

    @property
    def properties(self):
        process_node = self.fetch_process_node()
        # read the attributes directly instead of using the `get_list` method
        # to avoid error in case the `orm.List` object was converted to an `orm.Data`
        return (
            process_node.inputs.properties.base.attributes.get("list")
            if process_node
            else []
        )

    @property
    def outputs(self):
        process_node = self.fetch_process_node()
        return process_node.outputs if process_node else []

    def fetch_process_node(self) -> orm.ProcessNode | None:
        try:
            return orm.load_node(self.process_uuid) if self.process_uuid else None  # type: ignore
        except NotExistent:
            return None


class Confirmable(HasTraits):
    confirmed = tl.Bool(False)

    confirmation_exceptions = [
        "confirmed",
    ]

    def confirm(self):
        self.confirmed = True

    @tl.observe(tl.All)
    def _on_any_change(self, change):
        if change and change["name"] not in self.confirmation_exceptions:
            self._unconfirm()

    def _unconfirm(self):
        self.confirmed = False


class HasBlockers(HasTraits):
    blockers = tl.List(tl.Unicode())
    blocker_messages = tl.Unicode("")

    @property
    def is_blocked(self):
        return any(self.blockers)

    def update_blockers(self):
        blockers = list(self._check_blockers())
        if isinstance(self, HasModels):
            for _, model in self.get_models():
                if isinstance(model, HasBlockers):
                    blockers += model.blockers
        self.blockers = blockers

    def update_blocker_messages(self):
        if self.is_blocked:
            formatted = "\n".join(f"<li>{item}</li>" for item in self.blockers)
            self.blocker_messages = f"""
                <div class="alert alert-danger">
                    <b>The step is blocked due to the following reason(s):</b>
                    <ul>
                        {formatted}
                    </ul>
                </div>
            """
        else:
            self.blocker_messages = ""

    def _check_blockers(self):
        raise NotImplementedError



from __future__ import annotations

import typing as t
import traitlets as tl

from ..panel import Panel
from .model import SettingsModel

SM = t.TypeVar("SM", bound=SettingsModel)


class SettingsPanel(Panel[SM]):
    """Base model for settings panels."""

    updated = False

    def __init__(self, model: SM, **kwargs):
        super().__init__(model=model, **kwargs)
        self.links: list[tl.dlink | tl.link] = []



import traitlets as tl
from aiida import orm

from aiidalab_qe_base.mixins import HasModels
from aiidalab_qe_base.models import CodeModel

from ..settings import SettingsModel


class ResourceSettingsModel(SettingsModel, HasModels[CodeModel]):
    """Base model for resource setting models."""

    title = "Resources"
    identifier = "resources"

    global_codes = tl.Dict(
        key_trait=tl.Unicode(),
        value_trait=tl.Dict(),
    )

    warning_messages = tl.Unicode("")

    def __init__(self, *args, **kwargs):
        self.default_codes: dict[str, dict] = kwargs.pop("default_codes", {})

        super().__init__(*args, **kwargs)

        # Used by the code-setup thread to fetch code options
        self.DEFAULT_USER_EMAIL = orm.User.collection.get_default().email

    def add_model(self, identifier: str, model: CodeModel):
        super().add_model(identifier, model)
        code_key = model.default_calc_job_plugin.split(".")[-1]
        model.update(
            self.DEFAULT_USER_EMAIL,
            default_code=self.default_codes.get(code_key, {}).get("code"),
        )

    def refresh_codes(self):
        for _, code_model in self.get_models():
            code_key = code_model.default_calc_job_plugin.split(".")[-1]
            code_model.update(
                self.DEFAULT_USER_EMAIL,
                default_code=self.default_codes.get(code_key, {}).get("code"),
                refresh=True,
            )

    def get_model_state(self):
        return {
            "codes": {
                identifier: code_model.get_model_state()
                for identifier, code_model in self.get_models()
                if code_model.is_ready
            },
        }

    def set_model_state(self, parameters: dict):
        code_data = parameters.get("codes", {}) or self.default_codes
        for identifier, code_model in self.get_models():
            if identifier in code_data:
                code_model.set_model_state(code_data[identifier])

    def _check_blockers(self):
        return []



import os
import typing as t

from aiidalab_qe_base.mixins import HasInputStructure

from ..settings import SettingsPanel
from .model import ConfigurationSettingsModel

CSM = t.TypeVar("CSM", bound=ConfigurationSettingsModel)


class ConfigurationSettingsPanel(SettingsPanel[CSM]):
    """Base class for configuration settings panels."""

    def refresh(self, specific=""):
        """Refreshes the settings panel.

        Unlinks the panel's widgets. If the panel's model is included in the
        calculation, also updates the model's defaults. Resets the model to
        these defaults if there is no input structure.

        Parameters
        ----------
        `specific` : `str`, optional
            If provided, specifies the level of refresh.
        """
        self.updated = False
        self._unsubscribe()
        if self._model.include:
            self.update(specific)
        if "PYTEST_CURRENT_TEST" in os.environ:
            # Skip resetting to avoid having to inject a structure when testing
            return
        if isinstance(self._model, HasInputStructure) and not self._model.has_structure:
            self._reset()

    def update(self, specific=""):
        """Updates the model if not yet updated.

        Parameters
        ----------
        `specific` : `str`, optional
            If provided, specifies the level of update.
        """
        if self.updated:
            return
        if not self._model.loaded_from_process:
            self._model.update(specific)
        self.updated = True

    def _unsubscribe(self):
        """Unlinks any linked widgets."""
        for link in self.links:
            link.unlink()
        self.links.clear()

    def _reset(self):
        """Resets the model to present defaults."""
        self.updated = False
        self._model.reset()



from aiidalab_qe_base.models import Model


class PanelModel(Model):
    """Base class for all panel models.

    Attributes
    ----------
    `title` : `str`
        The title to be shown in the GUI.
    `identifier` : `str`
        Which plugin this panel belong to.
    """

    title = "Panel"
    identifier = "panel"
